/*
 * patchlevel.h --
 *
 * This file does nothing except define a "patch level" for Tcl.
 * The patch level is an integer that increments with each new
 * release or patch release.  It's used to make sure that Tcl
 * patches are applied in the correct order and only to appropriate
 * sources.
 */

#define TCL_PATCH_LEVEL 106
